@@ -5,6 +5,7 @@ import urllib.request
 from pathlib import Path
 from typing import Any, Dict, List, Optional, Tuple
 from urllib.parse import urlparse
+from os.path import basename
 
 from azure.identity import (ChainedTokenCredential, DefaultAzureCredential,
                             DeviceCodeCredential, EnvironmentCredential,
@@ -24,9 +25,9 @@ class _FeathrSynapseJobLauncher(SparkJobLauncher):
     """
     Submits spark jobs to a Synapse spark cluster.
     """
-    def __init__(self, synapse_dev_url: str, pool_name: str, datalake_dir: str, executor_size: str, executors: int):
+    def __init__(self, synapse_dev_url: str, pool_name: str, datalake_dir: str, executor_size: str, executors: int, credential = None):
         # use DeviceCodeCredential if EnvironmentCredential is not available
-        self.credential = DefaultAzureCredential()
+        self.credential = credential
         # use the same credential for authentication to avoid further login.
         self._api = _SynapseJobRunner(
             synapse_dev_url, pool_name, executor_size=executor_size, executors=executors, credential=self.credential)
@@ -50,8 +51,8 @@ class _FeathrSynapseJobLauncher(SparkJobLauncher):
 
         return self._datalake.download_file(result_path, local_folder)
 
-    def submit_feathr_job(self, job_name: str, main_jar_path: str,  main_class_name: str, arguments: List[str],
-                          reference_files_path: List[str], job_tags: Dict[str, str] = None,
+    def submit_feathr_job(self, job_name: str, main_jar_path: str = None,  main_class_name: str = None, arguments: List[str] = None,
+                          python_files: List[str]= None, reference_files_path: List[str] = None, job_tags: Dict[str, str] = None,
                           configuration: Dict[str, str] = None):
         """
         Submits the feathr job
@@ -74,7 +75,7 @@ class _FeathrSynapseJobLauncher(SparkJobLauncher):
             job_tags (str): tags of the job, for exmaple you might want to put your user ID, or a tag with a certain information
             configuration (Dict[str, str]): Additional configs for the spark job
         """
-
+        assert main_jar_path, 'main_jar_path should not be none or empty but it is none or empty.'
         if main_jar_path.startswith('abfs'):
             main_jar_cloud_path = main_jar_path
             logger.info(
@@ -84,7 +85,7 @@ class _FeathrSynapseJobLauncher(SparkJobLauncher):
                         main_jar_path, job_name)
             main_jar_cloud_path = self._datalake.upload_file_to_workdir(main_jar_path)
             logger.info('{} is uploaded to {} for running job: {}',
-                         main_jar_path, main_jar_cloud_path, job_name)
+                        main_jar_path, main_jar_cloud_path, job_name)
 
         reference_file_paths = []
         for file_path in reference_files_path:
@@ -94,6 +95,7 @@ class _FeathrSynapseJobLauncher(SparkJobLauncher):
         self.current_job_info = self._api.create_spark_batch_job(job_name=job_name,
                                                                  main_file=main_jar_cloud_path,
                                                                  class_name=main_class_name,
+                                                                 python_files=python_files,
                                                                  arguments=arguments,
                                                                  reference_files=reference_files_path,
                                                                  tags=job_tags,
@@ -135,8 +137,17 @@ class _FeathrSynapseJobLauncher(SparkJobLauncher):
             str: `output_path` field in the job tags
         """
         tags = self._api.get_spark_batch_job(self.current_job_info.id).tags
-        assert tags is not None
-        return tags[OUTPUT_PATH_TAG]
+        # in case users call this API even when there's no tags available
+        return None if tags is None else tags[OUTPUT_PATH_TAG]
+
+    def get_job_tags(self) -> Dict[str, str]:
+        """Get job tags
+
+        Returns:
+            Dict[str, str]: a dict of job tags
+        """
+        return self._api.get_spark_batch_job(self.current_job_info.id).tags
+
 
 
 class _SynapseJobRunner(object):
@@ -203,7 +214,7 @@ class _SynapseJobRunner(object):
         return self.client.spark_batch.cancel_spark_batch_job(job_id)
 
     def create_spark_batch_job(self, job_name, main_file, class_name=None,
-                               arguments=None,  reference_files=None, archives=None, configuration=None, tags=None):
+                               arguments=None, python_files=None, reference_files=None, archives=None, configuration=None, tags=None):
         """
         Submit a spark job to a certain cluster
         """
@@ -214,20 +225,19 @@ class _SynapseJobRunner(object):
         executor_cores = self.EXECUTOR_SIZE[self._executor_size]['Cores']
         executor_memory = self.EXECUTOR_SIZE[self._executor_size]['Memory']
 
-        # Adding spaces between brackets. This is to workaround this known YARN issue (when running Spark on YARN):
-        # https://issues.apache.org/jira/browse/SPARK-17814?focusedCommentId=15567964&page=com.atlassian.jira.plugin.system.issuetabpanels%3Acomment-tabpanel#comment-15567964
-        # updated_arguments = []
-        # for elem in arguments:
-        #     if type(elem) == str:
-        #         updated_arguments.append(elem.replace("}", " }"))
-        #     else:
-        #         updated_arguments.append(elem)
+        # need to put the jar in as dependencies for pyspark job
+        jars = jars + [main_file]
+
+        # If file=main_file, then it's using only Scala Spark
+        # If file=python_files[0], then it's using Pyspark
+        spark_execution_file = python_files[0] if python_files else main_file
 
         spark_batch_job_options = SparkBatchJobOptions(
             tags=tags,
             name=job_name,
-            file=main_file,
+            file=spark_execution_file,
             class_name=class_name,
+            python_files=python_files[1:],
             arguments=arguments,
             jars=jars,
             files=files,
@@ -259,7 +269,7 @@ class _DataLakeFiler(object):
         assert len(datalake_path_split) >= 3
 
         if credential is None:
-            credential = DefaultAzureCredential()
+            raise RuntimeError("Invalid credential provided.")
 
         account_url = "https://" + datalake_path_split[2]
 
@@ -278,7 +288,7 @@ class _DataLakeFiler(object):
             self.dir_client = self.file_system_client.get_directory_client('/')
 
         self.datalake_dir = datalake_dir + \
-            '/' if datalake_dir[-1] != '/' else datalake_dir
+                            '/' if datalake_dir[-1] != '/' else datalake_dir
 
     def upload_file_to_workdir(self, src_file_path: str) -> str:
         """
@@ -289,7 +299,7 @@ class _DataLakeFiler(object):
 
         src_parse_result = urlparse(src_file_path)
         if src_parse_result.scheme.startswith('http'):
-            file_name = os.path.basename(src_file_path)
+            file_name = basename(src_file_path)
             file_client = self.dir_client.create_file(file_name)
             # returned paths for the uploaded file
             returned_path = self.datalake_dir + file_name
@@ -315,7 +325,7 @@ class _DataLakeFiler(object):
         return returned_path
 
     def upload_file(self, src_file_path)-> str:
-        file_name = os.path.basename(src_file_path)
+        file_name = basename(src_file_path)
         logger.info("Uploading file {}", file_name)
         file_client = self.dir_client.create_file(file_name)
         returned_path = self.datalake_dir + file_name
@@ -327,7 +337,8 @@ class _DataLakeFiler(object):
 
     def download_file(self, target_adls_directory: str, local_dir_cache: str):
         """
-        Download file to a local cache
+        Download file to a local cache. Supporting download a folder and the content in its subfolder.
+        Note that the code will just download the content in the root folder, and the folder in the next level (rather than recursively for all layers of folders)
 
         Args:
             target_adls_directory (str): target ADLS directory
@@ -338,14 +349,40 @@ class _DataLakeFiler(object):
         parse_result = urlparse(target_adls_directory)
         directory_client = self.file_system_client.get_directory_client(
             parse_result.path)
+
         # returns the paths to all the files in the target director in ADLS
-        adls_paths = [file_path.name.split("/")[-1] for file_path in self.file_system_client.get_paths(
-            path=parse_result.path) if not file_path.is_directory][1:]
-        # need to generate list of local paths to write the files to
+        # get all the paths that are not under a directory
+        result_paths = [basename(file_path.name) for file_path in self.file_system_client.get_paths(
+            path=parse_result.path, recursive=False) if not file_path.is_directory]
+
+        # get all the paths that are directories and download them
+        result_folders = [file_path.name for file_path in self.file_system_client.get_paths(
+            path=parse_result.path) if file_path.is_directory]
+
+        # list all the files under the certain folder, and download them preserving the hierarchy
+        for folder in result_folders:
+            folder_name = basename(folder)
+            file_in_folder = [os.path.join(folder_name, basename(file_path.name)) for file_path in self.file_system_client.get_paths(
+            path=folder, recursive=False) if not file_path.is_directory]
+            local_paths = [os.path.join(local_dir_cache, file_name)
+                       for file_name in file_in_folder]
+            self._download_file_list(local_paths, file_in_folder, directory_client)
+
+        # download files that are in the result folder
         local_paths = [os.path.join(local_dir_cache, file_name)
-                       for file_name in adls_paths]
-        for idx, file_to_write in enumerate(tqdm(adls_paths,desc="Downloading result files: ")):
+                       for file_name in result_paths]
+        self._download_file_list(local_paths, result_paths, directory_client)
+
+        logger.info('Finish downloading files from {} to {}.',
+                    target_adls_directory,local_dir_cache)
+
+    def _download_file_list(self, local_paths: List[str], result_paths, directory_client):
+        '''
+        Download filelist to local
+        '''
+        for idx, file_to_write in enumerate(tqdm(result_paths,desc="Downloading result files: ")):
             try:
+                os.makedirs(os.path.dirname(local_paths[idx]), exist_ok=True)
                 local_file = open(local_paths[idx], 'wb')
                 file_client = directory_client.get_file_client(file_to_write)
                 download = file_client.download_file()
@@ -354,5 +391,3 @@ class _DataLakeFiler(object):
                 local_file.close()
             except Exception as e:
                 logger.error(e)
-        logger.info('Finish downloading files from {} to {}.',
-                    target_adls_directory,local_dir_cache)
